
RELEASE_VERSION='0.3.610538876'
BUILD_DATE='Thu Aug 11 15:23:14 UTC 2022'
BUILD_MACHINE_INFO='Linux runner-yxfxmkoy-project-28359933-concurrent-0cqtfz 5.4.181-99.354.amzn2.x86_64 #1 SMP Wed Mar 2 18:50:46 UTC 2022 x86_64 GNU/Linux'
GITLAB_CI_JOB_ID='2845584122'
GIT_COMMIT_SHA='ee4b93022a5aae3346ff3ebce2d4ed2dbe8ba82c'
GIT_BRANCH='main'
